(ns clj-in-act.ch2
  "The elements of Clojure: data structures and functions"
  (:require [clojure.repl :refer [doc find-doc apropos]]))


;;; 2.1 Coding at the REPL

(+ 1 2)                                                     ; Type this part at a REPL. Result is shown in comments.
;=> 3
(def my-addition (fn [operand1 operand2] (+ operand1 operand2)))
;=> #'clj-in-act.ch2/my-addition
(my-addition 1 2)
;=> 3
(my-addition 100 30)
;=> 130
(+ 1 2) "Two forms on one line!"
;=> 3
;=> "Two forms on one line!"


(println "Hello, world!")
;Hello, world!
;=> nil


"expression 1"
;=> "expression 1"
"expression 2"
;=> "expression 2"
*1
;=> "expression 2"
*3
;=> "expression 1"
"a" "b" "c"
;=> "a"
;=> "b"
;=> "c"
*3
;=> "a"
(def a-str *1)                                              ; We’ll save "a" for later.
;=> #'clj-in-act.ch2/a
a-str                                                       ; a-str contains the value that was in *1.
;=> "something else"
"something else"                                            ; This becomes the new value of *1.
;=> "something else"
a-str                                                       ; Still "a".
;=> "a"

;;; Examples of invalid forms

;())
;;=> ()
;;RuntimeException Unmatched delimiter: )  clojure.lang.Util.runtimeException (Util.java:221)

;*1
;;=> ()

;*e
;;=> #<ReaderException clojure.lang.LispReader$ReaderException: java.lang.RuntimeException: Unmatched delimiter: )>


;;; Looking up documentation
;;; In a REPL in the `user` namespace, `doc`, `find-doc` and `apropos` are
;;; available, but outside the `user` namespace you need to require these
;;; functions from the `clojure.repl` namespace. (We do this in the `ns` form
;;; at the top of this file.

(doc +)
;-------------------------
;clojure.core/+
;([] [x] [x y] [x y & more])
;Returns the sum of nums. (+) returns 0. Does not auto-promote
;longs, will throw on overflow. See also: +'


(+)                                                         ; No arguments: []
;=> 0
(+ 1)                                                       ; One argument: [x]
;=> 1
(+ 1 2)                                                     ; Two arguments: [x y]
;=> 3
(+ 1 2 3)                                                   ; Two or more arguments: [x y & more]
;=> 6
(+ 1 2 3 4 5 6 7 8)                                         ; Two or more arguments: [x y & more]
;=> 36



(find-doc "lazy")
;-------------------------
;clojure.string/split
;([s re] [s re limit])
;Splits string on a regular expression.  Optional argument limit is
;the maximum number of splits. Not lazy. Returns vector of the splits.
;-------------------------
;clojure.tools.nrepl/response-seq
;([transport] [transport timeout])
;Returns a lazy seq of messages received via the given Transport.
;Called with no further arguments, will block waiting for each message.
;The seq will end only when the underlying Transport is closed (i.e.
;returns nil from `recv`) or if a message takes longer than `timeout`
;millis to arrive.
;;;... more results omitted.
;=> nil


(apropos 'doc)
;=> (*remote-javadocs* add-remote-javadoc *local-javadocs* add-local-javadoc
;    javadoc doc find-doc)



;;; A few more points about Clojure syntax

(def x 1)
;=> #'clj-in-act.ch2/x
(cond
  (> x 0) "greater!"
  (= x 0) "zero!"
  (< x 0) "lesser!")
;=> "greater!"

;; Commas are whitespace
(+ 1 2 3 4 5)
;=> 15
(+ 1, 2, 3, 4, 5)
;=> 15
(+ 1, 2, 3, 4, 5)
;=> 15
(+ 1,,,,, 2, 3 4,, 5)
;=> 15


(def a-map {:a 1 :b 2 :c 3})
;=> #'clj-in-act.ch2/a-map
a-map
;=> {:c 3, :b 2, :a 1}


;;; Comments

;; This function does addition.
(defn add [x y]
  (+ x y))


(comment
  (defn this-is-not-working [x y]
    (+ x y)))
;=> nil


;; #_ is a reader macro which causes the next form to be completely ignored.

#_i-am-ignored
#_["me too"]


;;; 2.2 Clojure data structures

;;; Number type contagiousness

(+ 1 1N)
;=> 2N
(+ 1 1N 1/2)
;=> 5/2
(+ 1 1N 1/2 0.5M)
;=> 3.0M
(+ 1 1N 1/2 0.5M 0.5)
;=> 3.5


;;; Auto-promoting arithmetic functions

#_(inc 9223372036854775807)
;ArithmeticException integer overflow  clojure.lang.Numbers.throwIntOverflow (Numbers.java:1424)
(inc' 9223372036854775807)
;=> 9223372036854775808N


;;; Keywords and Symbols

(keyword "foo")
;=> :foo
(symbol "foo" "bar")
;=> foo/bar
(name :foo/bar)
;=> "bar"
(namespace :foo)                                            ; No namespace part returns `nil`.
;=> nil
(name "baz")                                                ; `name` returns strings unchanged.
;=> "baz"


;;; Lists

(list 1 2 3 4 5)
;=> (1 2 3 4 5)
(list? *1)
;=> true

(conj (list 1 2 3 4 5) 6)
;=> (6 1 2 3 4 5)

(conj (list 1 2 3) 4 5 6)
;=> (6 5 4 1 2 3)
(conj (conj (conj (list 1 2 3) 4) 5) 6)                     ; Same as above.
;=> (6 5 4 1 2 3)

(peek (list 1 2 3))
;=> 1
(pop (list 1 2 3))
;=> (2 3)
(peek (list))                                               ; The "head" of an empty list is `nil`.
;=> nil
#_(pop (list))                                              ; But the tail of an empty list is an exception!
;IllegalStateException Can't pop empty list  clojure.lang.PersistentList$EmptyList.pop (PersistentList.java:183)


(count (list))
;=> 0
(count (list 1 2 3 4))
;=> 4


#_(def three-numbers (1 2 3))
; CompilerException java.lang.ClassCastException: java.lang.Long cannot be cast to clojure.lang.IFn, compiling:(clj_in_act/ch2.clj:214:14)

(def three-numbers '(1 2 3))
;=> #'clj-in-act.ch2/three-numbers


;;; Vectors

(vector 10 20 30 40 50)
;=> [10 20 30 40 50]
(def the-vector [10 20 30 40 50])
;=> #'clj-in-act.ch2/the-vector

(get the-vector 2)
;=> 30
(nth the-vector 2)
;=> 30
(get the-vector 10)
;=> nil
#_(nth the-vector 10)
;IndexOutOfBoundsException   clojure.lang.PersistentVector.arrayFor (PersistentVector.java:107)

(assoc the-vector 2 25)                                     ; You can change an existing index,
;=> [10 20 25 40 50]
(assoc the-vector 5 60)                                     ; or add on to the end,
;=> [10 20 30 40 50 60]
#_(assoc the-vector 6 70)                                   ; But you cannot add past the end!
;IndexOutOfBoundsException   clojure.lang.PersistentVector.assocN (PersistentVector.java:137)


(conj [1 2 3 4 5] 6)
;=> [1 2 3 4 5 6]

(peek [1 2])
;=> 2
(pop [1 2])
;=> [1]
(peek [])                                                   ; Peeking an empty collection always returns nil.
;=> nil
#_(pop [])                                                  ; Popping an empty collection always throws.
;IllegalStateException Can't pop empty vector  clojure.lang.PersistentVector.pop (PersistentVector.java:381)

(the-vector 3)
;=> 40


;;; Maps

(def the-map {:a 1 :b 2 :c 3})
;=> #'clj-in-act.ch2/the-map

(hash-map :a 1 :b 2 :c 3)
;=> {:a 1, :c 3, :b 2}

(the-map :b)
;=> 2

(:b the-map)
;=> 2
(:z the-map 26)                                             ; Keywords can also return a default value if not found.
;=> 26


(def updated-map (assoc the-map :d 4))
;=> #'clj-in-act.ch2/updated-map
updated-map
;=> {:d 4, :a 1, :b 2, :c 3}
(dissoc updated-map :a)
;=> {:b 2, :c 3, :d 4}



(def users {:kyle {:date-joined "2009-01-01"
                   :summary     {:average {:monthly 1000
                                           :yearly  12000}}}})

(assoc-in users [:kyle :summary :average :monthly] 3000)
;=> {:kyle {:date-joined "2009-01-01", :summary {:average {:monthly 3000,
;                                                          :yearly 12000}}}}

(get-in users [:kyle :summary :average :monthly])
;=> 1000

(update-in users [:kyle :summary :average :monthly] + 500)
;=> {:kyle {:date-joined "2009-01-01", :summary {:average {:monthly 1500,
;                                                          :yearly 12000}}}}


;;; Sequences

(first (list 1 2 3))
;=> 1
(rest (list 1 2 3))
;=> (2 3)
(first [1 2 3])
;=> 1
(rest [1 2 3])
;=> (2 3)
(first {:a 1 :b 2})                                         ; Order of items is not guaranteed.
;=> [:b 2]
(rest {:a 1 :b 2})
;=> ([:a 1])
(first [])                                                  ; Empty collections return nil for first.
;=> nil
(rest [])                                                   ; Empty collections return an empty sequence for rest.
;=> ()

(cons 1 [2 3 4 5])
;=> (1 2 3 4 5)

(list? (cons 1 (list 2 3)))
;=> false

;;; 2.3 Program structure

(defn addition-function [x y]
  (+ x y))

;; Expanded form:
(def addition-function
  (fn [x y]
    (+ x y)))


;;; `let`

(defn average-pets []
  (/ (apply + (map :number-pets (vals users))) (count users)))


(defn average-pets []
  (let [user-data (vals users)
        pet-counts (map :number-pets user-data)
        ; apply calls a function with the items of a sequence as individual
        ; arguments. E.g., (apply + [1 2]) is the same as (+ 1 2).
        total (apply + pet-counts)]
    (/ total (count users))))


(let [x 1
      y 2
      z (+ x y)]
  z)
;=> 3


(defn average-pets []
  (let [user-data (vals users)
        pet-counts (map :number-pets user-data)
        value-from-println (println "total  pets:" pet-counts)
        total (apply + pet-counts)]
    (/ total (count users))))


(defn average-pets []
  (let [user-data (vals users)
        pet-counts (map :number-pets user-data)
        _ (println "total  pets:" pet-counts)
        total (apply + pet-counts)]
    (/ total (count users))))


;; `do` form
#_(if (is-something-true?)
  (do
    (log-message "in true branch")
    (store-something-in-db)
    (return-useful-value)))


;;; Program Flow

(if (> 5 2)
  "yes"
  "no")
;=> "yes"

(if-not (> 5 2) "yes" "no")
;=> "no"

(def x 1)
;=> #'clj-in-act.ch2/x
(cond
  (> x 0) "greater!"
  (= x 0) "zero!"
  :default "lesser!")
;=> "greater!"

(when (> 5 2)
  (println "five")
  (println "is")
  (println "greater")
  "done")
;five
;is
;greater
;=> "done"

(when-not (< 5 2)
  (println "two")
  (println "is")
  (println "smaller")
  "done")
;two
;is
;smaller
;=> "done"


;;; Logic functions

(and)
;=> true
(and :a :b :c)
;=> :c
(and :a nil :c)
;=> nil
(and :a false :c)
;=> false
;; In Clojure only nil and false are logically false. Everything else is true.
(and 0 "")
;=> ""

(or)
;=> nil
(or :a :b :c)
;=> :a
(or :a nil :c)
;=> :a
(or nil false)
;=> false
(or false nil)
;=> nil

(not true)
;=> false
(not 1)
;=> false
(not nil)
;=> true


;;; Numeric comparisons

(< 2 4 6 8)
;=> true

(< 2 4 3 8)
;=> false

(= 1 1N 1/1)
;=> true
(= 0.5 1/2)
;=> false
(= 0.5M 0.5)
;=> false
(= 0.5M 1/2)
;=> false


(== 1 1N 1/1)
;=> true
;; Numbers in different categories now compare roughly how you would expect.
(== 1/2 0.5M 0.5)
;=> true
1.9999999999999999
;=> 2.0

;; But == is not a magic wand against floating point precision and rounding behavior!
(== 2.0M 1.9999999999999999)
;=> true

;; All arguments must be numbers or Clojure will throw an exception.
#_(== :a 1)
;ClassCastException clojure.lang.Keyword cannot be cast to java.lang.Number  clojure.lang.Numbers.equiv (Numbers.java:206)
#_(== nil 1)
;NullPointerException clojure.lang.Numbers.ops (Numbers.java:961)

;(while true
;  "This line will be executed."
;  "The expression returns this string.")
;
;(while false
;  "None of these lines will be executed."
;  "The expression returns `nil`.")


(defn fact-loop [n]
  (loop [current n fact 1]
    (if (= current 1)
      ; Return whatever value fact ends up having when current is one.
      fact
      ; But if current is not one, repeat the loop with current reset to current
      ; minus one and with fact set to current times fact.
      (recur (dec current) (* fact current) ))))


;(defn fact-loop-invalid [n]
;  (loop [current n fact 1]
;    (if (= current 1)
;      fact
;      ;; This is an illegal position for recur.
;      (recur (dec current) (* fact current)))
;    (println "Done, current value:" current)))
;;CompilerException java.lang.UnsupportedOperationException: Can only recur from
;; tail position, compiling:(NO_SOURCE_PATH:5:7)

(defn run-report [user]
  (println "Running report for" user))

(defn dispatch-reporting-jobs [all-users]
  (doseq [user all-users]
    (run-report user)))


(dotimes [x 5]
  (println "X is" x))


(map inc [0 1 2 3])
;=> (1 2 3 4)

;; When supplying multiple sequences, each supplies an additional argument to the function.
(map + [0 1 2 3] [0 1 2 3])
;=> (0 2 4 6)

;; The length of the return value is the length of the shortest sequence.
(map + [0 1 2 3] [0 1 2])
;=> (0 2 4)


(defn non-zero-expenses [expenses]
  (let [non-zero? (fn [e] (not (zero? e)))]
    (filter non-zero? expenses)))
;=> #'clj-in-act.ch2/non-zero-expenses
;; Notice the 0 is gone from this sequence.
(non-zero-expenses [-2 -1 0 1 2 3])
;=> (-2 -1 1 2 3)


(defn non-zero-expenses [expenses]
  (remove zero? expenses))
;=> #'clj-in-act.ch2/non-zero-expenses

;; Notice we get the same result using remove, but we don’t need to create the
;; non-zero? function.
(non-zero-expenses [-2 -1 0 1 2 3])
;=> (-2 -1 1 2 3) #A


(defn factorial [n]
  (let [numbers (range 1 (+ n 1))]
    (reduce * numbers)))


(range 10)
;=> (0 1 2 3 4 5 6 7 8 9)

(factorial 5)
;=> 120

(defn factorial-steps [n]
  (let [numbers (range 1 (+ n 1))]
    (reductions * numbers)))
;=> #'clj-in-act.ch2/factorial-steps

;; Notice that the results of factorial-steps correspond to the results of
;; calling our normal factorial with all the values from 1 to 5.
(factorial-steps 5)
;=> (1 2 6 24 120)
(factorial 1)
;=> 1
(factorial 2)
;=> 2
(factorial 3)
;=> 6
(factorial 4)
;=> 24
(factorial 5)
;=> 120

;; Another practical illustration of map and range.
(map factorial (range 1 6))
;=> (1 2 6 24 120) #B


(def chessboard-labels
  (for [alpha "abcdefgh"
        num (range 1 9)]
    (str alpha num)))

chessboard-labels
;=> ("a1" "a2" "a3" "a4" "a5"  …  "h6" "h7" "h8")

(defn prime? [x]
  ;; Math/sqrt returns the square root of a number as a double.
  ;; The int is to truncate the result to an integer.
  (let [divisors (range 2 (inc (int (Math/sqrt x))))
        remainders (map (fn [d] (rem x d)) divisors)]
    (not (some zero? remainders))))


(defn primes-less-than [n]
  (for [x (range 2 (inc n))
        :when (prime? x)]
    x))

(primes-less-than 50)
;=> (2 3 5 7 11 13 17 19 23 29 31 37 41 43 47)


(defn pairs-for-primes [n]
  (let [z (range 2 (inc n))]
    (for [x z y z :when (prime? (+ x y))]
      (list x y))))

(pairs-for-primes 5)
;=> ((2 3) (2 5) (3 2) (3 4) (4 3) (5 2))


;;; Threading macros


(defn final-amount [principle rate time-periods]
  (* (Math/pow (+ 1 (/ rate 100)) time-periods) principle)) ;#A

(final-amount 100 20 1)
;=> 120.0
(final-amount 100 20 2)
;=> 144.0

(defn final-amount-> [principle rate time-periods]
  (-> rate
      (/ 100)
      (+ 1)
      (Math/pow time-periods)
      (* principle)))
;; Body expands to this:
;; (* (Math/pow (+ (/ rate 100) 1) time-periods) principle)


(final-amount-> 100 20 1)
;=> 120.0
(final-amount-> 100 20 2)
;=> 144.0

(defn factorial [n]
  (reduce * (range 1 (+ 1 n))))


(defn factorial->> [n]
  (->> n
       (+ 1)
       (range 1)
       (reduce *)))
;; Body expands to (reduce * (range 1 (+ 1 n)))

(factorial->> 5)
;=> 120


;; The first form is bound to the name <>. (You can use any name you want.)
(as-> {"a" [1 2 3 4]} <>
      (<> "a") ; Which can be used anywhere in the next form.
      (conj <> 10) ; <> is [1 2 3 4] before this form executes, and becomes [1 2 3 4 10] afterwards.
      (map inc <>)) ; inc increments a number by one.
;=> (2 3 4 5 11)

;; Expands to the following:

(let [<> {"a" [1 2 3 4]}
      <> (<> "a")
      <> (conj <> 10)
      <> (map inc <>)]
  <>)


(let [x 1 y 2]
  (cond-> []
          ;; The [] is only threaded through the right form. Conditions cannot
          ;; see the value being threaded.
          (odd? x)          (conj "x is odd")
          ;; The condition is false, so this form is skipped, but the threading
          ;; continues to the next pair of forms.
          (zero? (rem y 3)) (conj "y is divisible by 3")
          (even? y)         (conj "y is even")))
;=> ["x is odd" "y is even"]

;; Equivalent to cond->
(let [x 1 y 2]
  (as-> [] <>
        (if (odd? x)          (conj <> "x is odd")            <>)
        (if (zero? (rem y 3)) (conj <> "y is divisible by 3") <>)
        (if (even? y)         (conj <> "y is even")           <>)))
;=> ["x is odd" "y is even"]
