(ns clj-in-act.ch3.listings)

;;; Listing 3.1
;;; Mutually recursive functions that can blow the stack
(declare hat)
(defn cat [n]
  (when-not (zero? n)
    (when (zero? (rem n 100))
      (println "cat:" n))
    (hat (dec n))))

(defn hat [n]
  (when-not (zero? n)
    (if (zero? (rem n 100))
      (println "hat:" n))
    (cat (dec n))))


;;; Listing 3.2
;;; Mutually recursive functions that can be called with trampoline
(declare hatt)
(defn catt [n]
  (when-not (zero? n)
    (when (zero? (rem n 100))
      (println "catt:" n))
    (fn [] (hatt (dec n)))))

(defn hatt [n]
  (when-not (zero? n)
    (when (zero? (rem n 100))
      (println "hatt:" n))
    (fn [] (catt (dec n)))))


;;; Listing 3.3
;;; Function composition using higher-order functions
(def users
  [{:username     "kyle"
    :firstname    "Kyle"
    :lastname     "Smith"
    :balance      175.00M             ; Use BigDecimals for money!
    :member-since "2009-04-16"}
   {:username     "zak"
    :firstname    "Zackary"
    :lastname     "Jones"
    :balance      12.95M
    :member-since "2009-02-01"}
   {:username     "rob"
    :firstname    "Robert"
    :lastname     "Jones"
    :balance      98.50M
    :member-since "2009-03-30"}])
(defn sorter-using [ordering-fn]
  (fn [collection]
    (sort-by ordering-fn collection)))
(defn lastname-firstname [user]
  [(user :lastname) (user :firstname)])
(defn balance [user] (user :balance))
(defn username [user] (user :username))
(def poorest-first (sorter-using balance))
(def alphabetically (sorter-using username))
(def last-then-firstname (sorter-using lastname-firstname))


;;; Listing 3.4
;;; Dynamic scope in action
(def ^:dynamic *eval-me* 10)
(defn print-the-var [label]
  (println label *eval-me*))
(print-the-var "A:")
(binding [*eval-me* 20] ;; the first binding
  (print-the-var "B:")
  (binding [*eval-me* 30] ;; the second binding
    (print-the-var "C:"))
  (print-the-var "D:"))
(print-the-var "E:")


;;; Listing 3.5
;;; A higher-order function for aspect-oriented logging
(defn ^:dynamic twice [x]
  (println "original function")
  (* 2 x))

(defn call-twice [y]
  (twice y))

(defn with-log [function-to-call log-statement]
  (fn [& args]
    (println log-statement)
    (apply function-to-call args)))

(call-twice 10)

(binding [twice (with-log twice "Calling the twice function")]
  (call-twice 20))

(call-twice 30)


;;; Listing 3.6
;;; Using external libraries by calling use
(comment
  (ns org.currylogic.damages.http.expenses)
  (use 'clojure.data.json)
  (use 'clojure.xml)

  (declare load-totals)

  (defn import-transactions-xml-from-bank [url]
    (let [xml-document (parse url)]
      ;; more code here
      ))

  (defn totals-by-day [start-date end-date]
    (let [expenses-by-day (load-totals start-date end-date)]
      (json-str expenses-by-day))))


;;; Listing 3.7
;;; Using external libraries by calling require
(comment
  (ns org.currylogic.damages.http.expenses)
  (require '(clojure.data [json :as json-lib]))
  (require '(clojure [xml :as xml-core]))

  (declare load-totals)

  (defn import-transactions-xml-from-bank [url]
    (let [xml-document (xml-core/parse url)]
      ;; more code here
      ))

  (defn totals-by-day [start-date end-date]
    (let [expenses-by-day (load-totals start-date end-date)]
      (json-lib/json-str expenses-by-day))))


;;; Listing 3.8
;;; Using external libraries by calling require
(comment
  (ns org.currylogic.damages.http.expenses
    (:require [clojure.data.json :as json-lib]
              [clojure.xml :as xml-core]))

  (declare load-totals)

  (defn import-transactions-xml-from-bank [url]
    (let [xml-document (xml-core/parse url)]
      ;; more code here
      ))

  (defn totals-by-day [start-date end-date]
    (let [expenses-by-day (load-totals start-date end-date)]
      (json-lib/json-str expenses-by-day))))
