(ns clj-in-act.ch5
  "Exploring Clojure and Java interop"
  (:import (java.text SimpleDateFormat)
           (java.util Calendar)
           java.io.FileFilter))


;;; 5.1 Calling Java from Clojure

(def sdf (new SimpleDateFormat "yyyy-MM-dd"))
(def sdf (SimpleDateFormat. "yyyy-MM-dd"))


(. System getenv "PATH")
(. System (getenv "PATH"))

(import '(java.util Random))
;=> java.util.Random
(def rnd (Random. ))
;=> #'clj-in-act5/rnd
(. rnd nextInt 10)
;=> 4
(. rnd (nextInt 10))
;=> 3

(. Calendar DECEMBER)

(import '(java.util Calendar TimeZone))
;=> java.util.Calendar
(. (. (Calendar/getInstance) (getTimeZone)) (getDisplayName))
;=> "Pacific Standard Time" ; This may vary on your system.
(.. (Calendar/getInstance) (getTimeZone) (getDisplayName))
;=> "Pacific Standard Time"
(.. (Calendar/getInstance) getTimeZone getDisplayName)
;=> "Pacific Standard Time"
(.. (Calendar/getInstance)
    getTimeZone
    (getDisplayName true TimeZone/SHORT))
;=> "PDT"


(defn the-past-midnight-1 []
  (let [calendar-obj (Calendar/getInstance)]
    (.set calendar-obj Calendar/AM_PM Calendar/AM)
    (.set calendar-obj Calendar/HOUR 0)
    (.set calendar-obj Calendar/MINUTE 0)
    (.set calendar-obj Calendar/SECOND 0)
    (.set calendar-obj Calendar/MILLISECOND 0)
    (.getTime calendar-obj)))


(defn the-past-midnight-2 []
  (let [calendar-obj (Calendar/getInstance)]
    (doto calendar-obj
      (.set Calendar/AM_PM Calendar/AM)
      (.set Calendar/HOUR 0)
      (.set Calendar/MINUTE 0)
      (.set Calendar/SECOND 0)
      (.set Calendar/MILLISECOND 0))
    (.getTime calendar-obj)))

(map (fn [x] (.getBytes x)) ["amit" "rob" "kyle"])

(map #(.getBytes %) ["amit" "rob" "kyle"])

(map (memfn ^String getBytes) ["amit" "rob" "kyle"])

(.subSequence "Clojure" 2 5)
;=> "oju"

((memfn ^String subSequence ^Long start ^Long end) "Clojure" 2 5)
;=> "oju"

(bean (Calendar/getInstance))
;=> {:timeInMillis 1257466522295,
;    :minimalDaysInFirstWeek 1,
;    :lenient true,
;    :firstDayOfWeek 1,
;    :class java.util.GregorianCalendar
;    ;; other properties
;    }

(def tokens (.split "clojure.in.action" "\\."))

(alength tokens)
;=> 3
(aget tokens 2)
;=> "action"
(aset tokens 2 "actionable")
;=> "actionable"
(into [] tokens)
;=> ["clojure" "in" "actionable"]

(import 'java.awt.event.MouseAdapter)
(proxy [MouseAdapter] []
  (mousePressed [event]
    (println "Hey!")))


(reify java.io.FileFilter
  (accept [this f]
    (.isDirectory f)))

;;; 5.2 Compiling Clojure code to Java byte code

;;; If you are using leiningen, you can run the following commands from a shell.

;;; This compiles the AbstractClass
; $ lein javac

;;; This compiles the gen-class
; $ lein compile

;;; This runs the class
; $ lein run

;;; OR, if you don't have :main set in project.clj
; $ java -cp "$(lein classpath)" com.gentest.ConcreteClojureClass

;;; Output:
; Init value: READY
;Constructor: a
;getCurrentStatus from - com.gentest.ConcreteClojureClass
;The Secret
;READY
;Init values: READY , SET
;Constructor: a, b
;SET
;In run!
;I'm a com.gentest.ConcreteClojureClass
;GO
